"""Process input or output fields."""
import collections
import json

import resolwe_runtime_utils


class ValidationError(Exception):
    """Field value validation error."""


class Field:
    """Process input or output field."""

    field_type = None

    def __init__(
        self,
        label=None,
        required=True,
        description=None,
        default=None,
        choices=None,
        hidden=False,
    ):
        """Construct a field descriptor."""
        self.name = None
        self.process = None
        self.label = label
        self.required = required
        self.description = description
        self.default = default
        self.choices = choices
        self.hidden = hidden

    def get_field_type(self):
        """Return this field's type."""
        return self.field_type

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        self.name = name
        self.process = process
        fields[name] = self

    def to_python(self, value):
        """Convert value if needed."""
        return value

    def to_schema(self):
        """Return field schema for this field."""
        if not self.name or not self.process:
            raise ValueError("field is not registered with process")

        schema = {
            "name": self.name,
            "type": self.get_field_type(),
        }
        if self.required is not None:
            schema["required"] = self.required
        if self.label is not None:
            schema["label"] = self.label
        if self.description is not None:
            schema["description"] = self.description
        if self.default is not None:
            schema["default"] = self.default
        if self.hidden is not None:
            schema["hidden"] = self.hidden
        if self.choices is not None:
            for choice, label in self.choices:
                schema.setdefault("choices", []).append(
                    {"label": label, "value": choice,}
                )

        return schema

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save(self.name, value))

    def validate(self, value):
        """Validate field value."""
        if self.required and value is None:
            raise ValidationError("field is required")

        if value is not None and self.choices is not None:
            choices = [choice for choice, _ in self.choices]
            if value not in choices:
                raise ValidationError(
                    "field must be one of: {}".format(", ".join(choices),)
                )

    def clean(self, value):
        """Run validators and return the clean value."""
        if value is None:
            value = self.default

        try:
            value = self.to_python(value)
            self.validate(value)
        except ValidationError as error:
            raise ValidationError(
                "invalid value for {}: {}".format(self.name, error.args[0])
            )
        return value

    def __repr__(self):
        """Return string representation."""
        return '<{klass} name={name} type={type} label="{label}">'.format(
            klass=self.__class__.__name__,
            name=self.name,
            type=self.get_field_type(),
            label=self.label,
        )


class StringField(Field):
    """String field."""

    field_type = "basic:string"

    def validate(self, value):
        """Validate field value."""
        if value is not None and not isinstance(value, str):
            raise ValidationError("field must be a string")

        super().validate(value)


class TextField(StringField):
    """Text field."""

    field_type = "basic:text"


class BooleanField(Field):
    """Boolean field."""

    field_type = "basic:boolean"

    def validate(self, value):
        """Validate field value."""
        if value is not None and not isinstance(value, bool):
            raise ValidationError("field must be a boolean")

        super().validate(value)


class IntegerField(Field):
    """Integer field."""

    field_type = "basic:integer"

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            try:
                return int(value)
            except (TypeError, ValueError):
                raise ValidationError("field must be an integer")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save(self.name, str(value)))


class FloatField(Field):
    """Float field."""

    # TODO: Fix the underlying field into basic:float once that is renamed.
    field_type = "basic:decimal"

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            try:
                return float(value)
            except (TypeError, ValueError):
                raise ValidationError("field must be a float")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save(self.name, str(value)))


class DateField(Field):
    """Date field."""

    field_type = "basic:date"


class DateTimeField(Field):
    """Date time field."""

    field_type = "basic:datetime"


class UrlField(Field):
    """URL field."""

    # Url types.
    DOWNLOAD = "download"
    VIEW = "view"
    LINK = "link"

    URL_TYPES = (DOWNLOAD, VIEW, LINK)

    def __init__(self, url_type, *args, **kwargs):
        """Construct an URL field descriptor.

        :param url_type: Type of URL
        """
        if url_type not in self.URL_TYPES:
            raise ValueError(
                "url_type must be one of: {}".format(", ".join(self.URL_TYPES))
            )

        self.url_type = url_type
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, str):
            return value
        elif isinstance(value, dict):
            try:
                value = value["url"]
            except KeyError:
                raise ValidationError("dictionary must contain an 'url' element")

            if not isinstance(value, str):
                raise ValidationError("field's url element must be a string")

            return value
        elif not isinstance(value, None):
            raise ValidationError("field must be a string or a dict")

    def get_field_type(self):
        """Return this field's type."""
        return "basic:url:{}".format(self.url_type)


class SecretField(Field):
    """Secret field."""

    field_type = "basic:secret"


class FileDescriptor:
    """Descriptor for accessing files."""

    def __init__(
        self,
        path,
        size=None,
        total_size=None,
        is_remote=False,
        file_temp=None,
        refs=None,
    ):
        """Construct a file descriptor."""
        self.path = path
        self.size = size
        self.total_size = total_size
        self.is_remote = is_remote
        self.file_temp = file_temp
        if refs is None:
            refs = []
        self.refs = refs

    def import_file(self, imported_format=None, progress_from=0.0, progress_to=None):
        """Import field source file to working directory.

        :param imported_format: Import file format (extracted, compressed or both)
        :param progress_from: Initial progress value
        :param progress_to: Final progress value
        :return: Destination file path (if extracted and compressed, extracted path given)
        """
        if not hasattr(resolwe_runtime_utils, "import_file"):
            raise RuntimeError("Requires resolwe-runtime-utils >= 2.0.0")

        if imported_format is None:
            imported_format = resolwe_runtime_utils.ImportedFormat.BOTH

        return resolwe_runtime_utils.import_file(
            src=self.file_temp,
            file_name=self.path,
            imported_format=imported_format,
            progress_from=progress_from,
            progress_to=progress_to,
        )

    def __repr__(self):
        """Return string representation."""
        return "<FileDescriptor path={}>".format(self.path)


class FileField(Field):
    """File field."""

    field_type = "basic:file"

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, FileDescriptor):
            return value
        elif isinstance(value, str):
            return FileDescriptor(value)
        elif isinstance(value, dict):
            try:
                path = value["file"]
            except KeyError:
                raise ValidationError("dictionary must contain a 'file' element")

            if not isinstance(path, str):
                raise ValidationError("field's file element must be a string")

            size = value.get("size", None)
            if size is not None and not isinstance(size, int):
                raise ValidationError("field's size element must be an integer")

            total_size = value.get("total_size", None)
            if total_size is not None and not isinstance(total_size, int):
                raise ValidationError("field's total_size element must be an integer")

            is_remote = value.get("is_remote", None)
            if is_remote is not None and not isinstance(is_remote, bool):
                raise ValidationError("field's is_remote element must be a boolean")

            file_temp = value.get("file_temp", None)
            if file_temp is not None and not isinstance(file_temp, str):
                raise ValidationError("field's file_temp element must be a string")

            refs = value.get("refs", None)
            if refs is not None and not isinstance(refs, list):
                # TODO: Validate that all refs are strings.
                raise ValidationError("field's refs element must be a list of strings")

            return FileDescriptor(
                path,
                size=size,
                total_size=total_size,
                is_remote=is_remote,
                file_temp=file_temp,
                refs=refs,
            )
        elif not isinstance(value, None):
            raise ValidationError("field must be a FileDescriptor, string or a dict")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(
            resolwe_runtime_utils.save_file(self.name, value.path, *value.refs)
        )


class FileHtmlField(FileField):
    """HTML file field."""

    field_type = "basic:file:html"


class DirDescriptor:
    """Descriptor for accessing directories."""

    def __init__(self, path, size=None, total_size=None, refs=None):
        """Construct a file descriptor."""
        self.path = path
        self.size = size
        self.total_size = total_size
        if refs is None:
            refs = []
        self.refs = refs

    def __repr__(self):
        """Return string representation."""
        return "<DirDescriptor path={}>".format(self.path)


class DirField(Field):
    """Directory field."""

    field_type = "basic:dir"

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, DirDescriptor):
            return value
        elif isinstance(value, str):
            return DirDescriptor(value)
        elif isinstance(value, dict):
            try:
                path = value["dir"]
            except KeyError:
                raise ValidationError("dictionary must contain a 'dir' element")

            if not isinstance(path, str):
                raise ValidationError("field's dir element must be a string")

            size = value.get("size", None)
            if size is not None and not isinstance(size, int):
                raise ValidationError("field's size element must be an integer")

            total_size = value.get("total_size", None)
            if total_size is not None and not isinstance(total_size, int):
                raise ValidationError("field's total_size element must be an integer")

            refs = value.get("refs", None)
            if refs is not None and not isinstance(refs, list):
                # TODO: Validate that all refs are strings.
                raise ValidationError("field's refs element must be a list of strings")

            return DirDescriptor(path, size=size, total_size=total_size, refs=refs,)
        elif not isinstance(value, None):
            raise ValidationError("field must be a DirDescriptor, string or a dict")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(
            resolwe_runtime_utils.save_dir(self.name, value.path, *value.refs)
        )


class JsonField(Field):
    """JSON field."""

    field_type = "basic:json"

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            return json.loads(json.dumps(value))


class ListField(Field):
    """Generic list field."""

    def __init__(self, inner, *args, **kwargs):
        """Construct a list field."""
        if not isinstance(inner, Field):
            raise TypeError("inner field must be an instance of Field")

        self.inner = inner
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        super().contribute_to_class(process, fields, name)

        self.inner.name = name
        self.inner.process = process

    def to_python(self, value):
        """Convert value if needed."""
        return [self.inner.to_python(v) for v in value]

    def to_output(self, value):
        """Convert value to process output format."""
        return {self.name: [self.inner.to_output(v)[self.name] for v in value]}

    def get_field_type(self):
        """Return this field's type."""
        return "list:{}".format(self.inner.get_field_type())

    def validate(self, value):
        """Validate field value."""
        if value is not None:
            if not isinstance(value, list):
                raise ValidationError("field must be a list")

            for index, element in enumerate(value):
                try:
                    self.inner.validate(element)
                except ValidationError as error:
                    raise ValidationError(
                        "invalid element {}: {}".format(index, error.args[0],)
                    )

        super().validate(value)


class DataDescriptor:
    """Descriptor for accessing data objects."""

    def __init__(self, data_id, field, cache):
        """Construct a data descriptor.

        :param data_id: Data object primary key
        :param field: Field this descriptor is for
        :param cache: Optional cached object to use
        """
        super().__setattr__("_data_id", data_id)

        # Map output fields to a valid Python process syntax.
        for field_descriptor in cache["__output_schema"]:
            field_name = field_descriptor["name"]

            if field_name not in cache:
                # Non-required fields may be missing.
                cache[field_name] = None
                continue

            field_type = field_descriptor["type"].rstrip(":")

            if field_type.startswith("list:"):
                field_class = ALL_FIELDS_MAP[field_type[len("list:") :]]
                extra_kwargs = {}
                if issubclass(field_class, DataField):
                    extra_kwargs["data_type"] = field_type[len("list:data:") :]
                field = ListField(field_class(**extra_kwargs))
            else:
                field_class = ALL_FIELDS_MAP[field_type]
                extra_kwargs = {}
                if issubclass(field_class, DataField):
                    extra_kwargs["data_type"] = field_type[len("data:") :]
                field = field_class(**extra_kwargs)

            value = cache[field_name]
            value = field.clean(value)
            cache[field_name] = value

        super().__setattr__("_cache", cache)

    def _populate_cache(self):
        """Fetch data object from the backend if needed."""
        if self._data_id is None:
            return

        if self._cache is None:
            # TODO: Implement fetching via the protocol once available.
            raise NotImplementedError

    def _get(self, key):
        """Return given key from cache."""
        self._populate_cache()
        if key not in self._cache:
            raise AttributeError("DataField has no member {}".format(key))

        return self._cache[key]

    @property
    def id(self):
        """Primary key of this data object."""
        return self._data_id

    @property
    def type(self):
        """Type of this data object."""
        return self._get("__type")

    @property
    def descriptor(self):
        """Descriptor of this data object."""
        return self._get("__descriptor")

    @property
    def name(self):
        """Get this data name."""
        return self._get("__name")

    @property
    def entity_name(self):
        """Entity name."""
        return self._get("__entity_name")

    def __getattr__(self, key):
        """Get attribute."""
        return self._get(key)

    def __setattr__(self, key, value):
        """Set attribute."""
        raise AttributeError("inputs are read-only")

    def __repr__(self):
        """Return string representation."""
        return "<DataDescriptor id={}>".format(self._data_id)


class DataField(Field):
    """Data object field."""

    field_type = "data"

    def __init__(self, data_type, *args, **kwargs):
        """Construct a data field."""
        # TODO: Validate data type format.
        self.data_type = data_type
        super().__init__(*args, **kwargs)

    def get_field_type(self):
        """Return this field's type."""
        return "data:{}".format(self.data_type)

    def to_python(self, value):
        """Convert value if needed."""
        cache = None
        if value is None:
            return None

        if isinstance(value, DataDescriptor):
            return value
        elif isinstance(value, dict):
            # Allow pre-hydrated data objects.
            cache = value
            try:
                value = cache["__id"]
            except KeyError:
                raise ValidationError("dictionary must contain an '__id' element")
        else:
            raise ValidationError("field must be a DataDescriptor or dict")

        return DataDescriptor(value, self, cache)


class GroupDescriptor:
    """Group descriptor."""

    def __init__(self, value):
        """Construct a group descriptor."""
        self._value = value

    def __getattr__(self, name):
        """Get attribute."""
        try:
            return self._value[name]
        except KeyError:
            raise AttributeError(name)


class GroupField(Field):
    """Group field."""

    field_type = "basic:group"

    def __init__(
        self,
        field_group,
        label=None,
        description=None,
        disabled=False,
        collapsed=False,
        hidden=False,
    ):
        """Construct a group field."""
        super().__init__(
            label=label, required=None, description=description, hidden=hidden,
        )

        self.disabled = disabled
        self.collapsed = collapsed
        self.field_group = field_group
        self.fields = collections.OrderedDict()

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        # Use order-preserving definition namespace (__dict__) to respect the
        # order of GroupField's fields definition.
        for field_name in self.field_group.__dict__:
            if field_name.startswith("_"):
                continue

            field = getattr(self.field_group, field_name)
            field.contribute_to_class(process, self.fields, field_name)

        super().contribute_to_class(process, fields, name)

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, GroupDescriptor):
            value = value._value

        result = {}
        for name, field in self.fields.items():
            result[name] = field.to_python(value.get(name, None))

        return GroupDescriptor(result)

    def to_schema(self):
        """Return field schema for this field."""
        schema = super().to_schema()
        if self.disabled is not None:
            schema["disabled"] = self.disabled
        if self.collapsed is not None:
            schema["collapsed"] = self.collapsed

        group = []
        for field in self.fields.values():
            group.append(field.to_schema())
        schema["group"] = group

        return schema


# List of available fields.
ALL_FIELDS = [
    StringField,
    TextField,
    BooleanField,
    IntegerField,
    FloatField,
    DateField,
    DateTimeField,
    UrlField,
    SecretField,
    FileField,
    FileHtmlField,
    DirField,
    JsonField,
    ListField,
    DataField,
    GroupField,
]

ALL_FIELDS_MAP = {field.field_type: field for field in ALL_FIELDS}


def get_available_fields():
    """Return a list of available field classes."""
    return ALL_FIELDS



"""Data viewset."""
from elasticsearch_dsl.query import Q

from rest_framework import exceptions, mixins, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.elastic.composer import composer
from resolwe.elastic.viewsets import ElasticSearchCombinedViewSet
from resolwe.flow.models import Data, Process
from resolwe.flow.models.utils import fill_with_defaults
from resolwe.flow.serializers import DataSerializer
from resolwe.flow.utils import get_data_checksum
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.shortcuts import get_objects_for_user

from ..elastic_indexes import DataDocument
from .mixins import (
    ParametersMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)


class DataViewSet(
    ElasticSearchCombinedViewSet,
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.DestroyModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ParametersMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`Data` objects."""

    queryset = Data.objects.all().prefetch_related(
        "process", "descriptor_schema", "contributor", "collection", "entity"
    )
    serializer_class = DataSerializer
    permission_classes = (get_permissions_class(),)
    document_class = DataDocument

    filtering_fields = (
        "id",
        "slug",
        "version",
        "name",
        "name_contains",
        "created",
        "modified",
        "contributor",
        "contributor_name",
        "owners",
        "owners_name",
        "status",
        "process",
        "process_type",
        "type",
        "process_name",
        "tags",
        "collection",
        "entity",
        "started",
        "finished",
        "text",
        "process_slug",
    )
    filtering_map = {
        "name": "name.raw",
        "name_contains": "name.ngrams",
        "contributor": "contributor_id",
        "contributor_name": "contributor_name.ngrams",
        "owners": "owner_ids",
        "owners_name": "owner_names.ngrams",
        "process_name": "process_name.ngrams",
    }
    ordering_fields = (
        "id",
        "created",
        "modified",
        "started",
        "finished",
        "name",
        "contributor",
        "process_name",
        "process_type",
        "type",
    )
    ordering_map = {
        "name": "name.raw",
        "process_type": "process_type.raw",
        "type": "type.raw",
        "process_name": "process_name.raw",
        "contributor": "contributor_sort",
    }
    ordering = "-created"

    def custom_filter_tags(self, value, search):
        """Support tags query."""
        if not isinstance(value, list):
            value = value.split(",")

        filters = [Q("match", **{"tags": item}) for item in value]
        search = search.query("bool", must=filters)

        return search

    def custom_filter_text(self, value, search):
        """Support general query using the 'text' attribute."""
        if isinstance(value, list):
            value = " ".join(value)

        should = [
            Q("match", slug={"query": value, "operator": "and", "boost": 10.0}),
            Q(
                "match",
                **{"slug.ngrams": {"query": value, "operator": "and", "boost": 5.0}},
            ),
            Q("match", name={"query": value, "operator": "and", "boost": 10.0}),
            Q(
                "match",
                **{"name.ngrams": {"query": value, "operator": "and", "boost": 5.0}},
            ),
            Q(
                "match",
                contributor_name={"query": value, "operator": "and", "boost": 5.0},
            ),
            Q(
                "match",
                **{
                    "contributor_name.ngrams": {
                        "query": value,
                        "operator": "and",
                        "boost": 2.0,
                    }
                },
            ),
            Q("match", owner_names={"query": value, "operator": "and", "boost": 5.0}),
            Q(
                "match",
                **{
                    "owner_names.ngrams": {
                        "query": value,
                        "operator": "and",
                        "boost": 2.0,
                    }
                },
            ),
            Q("match", process_name={"query": value, "operator": "and", "boost": 5.0}),
            Q(
                "match",
                **{
                    "process_name.ngrams": {
                        "query": value,
                        "operator": "and",
                        "boost": 2.0,
                    }
                },
            ),
            Q("match", status={"query": value, "operator": "and", "boost": 2.0}),
            Q("match", type={"query": value, "operator": "and", "boost": 2.0}),
        ]

        # Add registered text extensions.
        for extension in composer.get_extensions(self):
            if hasattr(extension, "text_filter"):
                should += extension.text_filter(value)

        search = search.query("bool", should=should)

        return search

    @action(detail=False, methods=["post"])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Data`` objects."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        ids = self.get_ids(request.data)
        queryset = get_objects_for_user(
            request.user, "view_data", Data.objects.filter(id__in=ids)
        )
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Data objects with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        # TODO support ``inherit_collection``
        duplicated = queryset.duplicate(contributor=request.user)

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)

    @action(detail=False, methods=["post"])
    def get_or_create(self, request, *args, **kwargs):
        """Get ``Data`` object if similar already exists, otherwise create it."""
        response = self.perform_get_or_create(request, *args, **kwargs)
        if response:
            return response

        return super().create(request, *args, **kwargs)

    def perform_get_or_create(self, request, *args, **kwargs):
        """Perform "get_or_create" - return existing object if found."""
        self.define_contributor(request)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        process = serializer.validated_data.get("process")
        process_input = request.data.get("input", {})

        fill_with_defaults(process_input, process.input_schema)

        checksum = get_data_checksum(process_input, process.slug, process.version)
        data_qs = Data.objects.filter(
            checksum=checksum,
            process__persistence__in=[
                Process.PERSISTENCE_CACHED,
                Process.PERSISTENCE_TEMP,
            ],
        )
        data_qs = get_objects_for_user(request.user, "view_data", data_qs)
        if data_qs.exists():
            data = data_qs.order_by("created").last()
            serializer = self.get_serializer(data)
            return Response(serializer.data)

    def _parents_children(self, request, queryset):
        """Process given queryset and return serialized objects."""
        queryset = get_objects_for_user(request.user, "view_data", queryset)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    @action(detail=True)
    def parents(self, request, pk=None):
        """Return parents of the current data object."""
        return self._parents_children(request, self.get_object().parents)

    @action(detail=True)
    def children(self, request, pk=None):
        """Return children of the current data object."""
        return self._parents_children(request, self.get_object().children)



"""Constants used in Django Channels."""

# Channel used for notifying workers from ORM signals.
CHANNEL_PURGE_WORKER = "flow.purge"
# Message type for starting the Data purge.
TYPE_PURGE_RUN = "purge.run"



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-03-18 11:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="TestDependency",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
            ],
        ),
        migrations.CreateModel(
            name="TestModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                ("field_process_type", models.CharField(max_length=100)),
                ("number", models.IntegerField()),
                ("date", models.DateTimeField(null=True)),
            ],
            options={
                "default_permissions": (),
                "permissions": (
                    ("view_testmodel", "Can view model"),
                    ("edit_testmodel", "Can edit model"),
                ),
            },
        ),
        migrations.CreateModel(
            name="TestModelWithDependency",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                ("dependencies", models.ManyToManyField(to="test_app.TestDependency")),
                (
                    "dependency",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="main_dep",
                        to="test_app.TestDependency",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="TestSelfDependency",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                (
                    "dependencies",
                    models.ManyToManyField(
                        to="test_app.TestSelfDependency",
                        symmetrical=False,
                        related_name="parents",
                    ),
                ),
            ],
        ),
    ]



# pylint: disable=missing-docstring
import builtins
import copy
import gzip
import io
import json
import os.path
import shutil
import tarfile
import tempfile
import unittest.mock as mock

from resolwe.flow.models import Storage
from resolwe.test import ProcessTestCase, TestCase, is_testing, tag_process

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")
TEST_FILES_DIR = os.path.join(os.path.dirname(__file__), "files")


class TestingFrameworkTestCase(TestCase):
    def setUp(self):
        self.dummy_case = ProcessTestCase.__new__(ProcessTestCase)
        self.dummy_case.files_path = ""
        self.dummy_case._debug_info = lambda _: ""
        setattr(self.dummy_case, "assertEqual", self.assertEqual)

        self.temp_dir = tempfile.mkdtemp()
        open(os.path.join(self.temp_dir, "extensionless file"), "w").close()
        open(os.path.join(self.temp_dir, "text file.txt"), "w").close()
        os.mkdir(os.path.join(self.temp_dir, "dir"))
        open(os.path.join(self.temp_dir, "dir", "another file"), "w").close()
        os.mkdir(os.path.join(self.temp_dir, "dir", "empty dir.ext"))

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    @mock.patch("os.path.isfile")
    def test_assert_file_date_nofilter(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
        output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, "open", open_mock):
            obj_mock = mock.MagicMock()
            self.assertRaises(
                AssertionError,
                ProcessTestCase._assert_file,
                self.dummy_case,
                obj_mock,
                "",
                "",
            )

    @mock.patch("os.path.isfile")
    def test_assert_file_date_filter(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
        output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, "open", open_mock):
            obj_mock = mock.MagicMock()

            def date_in_line(line):
                return line.startswith(b"date")

            ProcessTestCase._assert_file(
                self.dummy_case, obj_mock, "", "", file_filter=date_in_line
            )

    @mock.patch("os.path.isfile")
    def test_assert_file_sort_false(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"A\nB\nC\n")
        output2_file = io.BytesIO(b"A\nC\nB\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, "open", open_mock):
            obj_mock = mock.MagicMock()
            self.assertRaises(
                AssertionError,
                ProcessTestCase._assert_file,
                self.dummy_case,
                obj_mock,
                "",
                "",
            )

    @mock.patch("os.path.isfile")
    def test_assert_file_sort_true(self, isfile_mock):
        isfile_mock.return_value = True
        output1_file = io.BytesIO(b"A\nB\nC\n")
        output2_file = io.BytesIO(b"A\nC\nB\n")
        open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
        with mock.patch.object(builtins, "open", open_mock):
            obj_mock = mock.MagicMock()
            ProcessTestCase._assert_file(self.dummy_case, obj_mock, "", "", sort=True)

    def test_assert_json_storage_object(self):
        example_json = {"foo": [1.0, 2.5, 3.14], "bar": ["ba", "cd"]}

        obj_mock = mock.MagicMock()

        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = "no_id"
        storage_mock.json = example_json

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        with gzip.open(gzipped_json_file, mode="wt") as f:
            json.dump(example_json, f)
        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)

        join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
        isfile_mock = mock.MagicMock(return_value=True)
        with mock.patch.object(os.path, "join", join_mock):
            with mock.patch.object(os.path, "isfile", isfile_mock):
                ProcessTestCase.assertJSON(
                    self.dummy_case, obj_mock, storage_mock, "", "foo.gz"
                )

    @mock.patch("resolwe.flow.models.Storage.objects.get")
    def test_assert_json_storage_id(self, get_mock):
        example_json = {"foo": [1.0, 2.5, 3.14], "bar": ["ba", "cd"]}

        obj_mock = mock.MagicMock()

        storage_id = "no_id"
        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = storage_id
        storage_mock.json = example_json
        get_mock.side_effect = [storage_mock]

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        with gzip.open(gzipped_json_file, mode="wt") as f:
            json.dump(example_json, f)
        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)

        join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
        isfile_mock = mock.MagicMock(return_value=True)
        with mock.patch.object(os.path, "join", join_mock):
            with mock.patch.object(os.path, "isfile", isfile_mock):
                ProcessTestCase.assertJSON(
                    self.dummy_case, obj_mock, storage_id, "", "foo.gz"
                )

    def test_assert_json_file_missing(self):
        example_json = {"foo": [1.0, 2.5, 3.14], "bar": ["ba", "cd"]}

        obj_mock = mock.MagicMock()

        storage_mock = mock.MagicMock(spec=Storage)
        storage_mock.id = "no_id"
        storage_mock.json = example_json

        # use in-memory binary stream object for speed and simplicity
        gzipped_json_file = io.BytesIO()
        join_modified_values = [gzipped_json_file]
        orig_join = os.path.join

        # NOTE: coverage tool needs original 'os.path.join'
        def join_side_effect(path, *paths):
            if join_modified_values:
                return join_modified_values.pop(0)
            else:
                return orig_join(path, *paths)

        join_mock = mock.MagicMock(side_effect=join_side_effect)
        isfile_mock = mock.MagicMock(return_value=False)
        with mock.patch.object(os.path, "join", join_mock):
            with mock.patch.object(os.path, "isfile", isfile_mock):

                self.assertRaisesRegex(
                    AssertionError,
                    "Output file .* missing so it was created.",
                    ProcessTestCase.assertJSON,
                    self.dummy_case,
                    obj_mock,
                    storage_mock,
                    "",
                    "foo.gz",
                )

        # set seek position of the binary stream object back to 0
        gzipped_json_file.seek(0)
        with gzip.open(gzipped_json_file, mode="rt") as f:
            unzipped_json = json.load(f)
        self.assertEqual(example_json, unzipped_json)

    def test_debug_info_non_ascii(self):
        non_ascii_text = "Some non-ascii chars č ü €"

        dummy_case = ProcessTestCase.__new__(ProcessTestCase)

        obj_mock = mock.MagicMock()
        obj_mock.pk = "no_id"

        with tempfile.NamedTemporaryFile(mode="wb") as stdout_file:
            stdout_file.write(non_ascii_text.encode("utf-8"))
            # set seek position of the file back to 0
            stdout_file.seek(0)
            with mock.patch.object(os.path, "join", side_effect=[stdout_file.name]):
                with mock.patch.object(os.path, "isfile", return_value=True):

                    # https://github.com/PyCQA/pylint/issues/1653
                    self.assertRegex(
                        ProcessTestCase._debug_info(dummy_case, obj_mock),
                        non_ascii_text,
                    )

    def test_assert_almost_equal_generic(self):
        self.assertAlmostEqualGeneric(1.00000001, 1.0)
        self.assertAlmostEqualGeneric([1.00000001], [1.0])
        self.assertAlmostEqualGeneric({"foo": 1.00000001}, {"foo": 1.0})
        self.assertAlmostEqualGeneric({"foo": [1.00000001]}, {"foo": [1.0]})
        self.assertAlmostEqualGeneric(
            {"foo": 1.00000001, "bar": "moo"}, {"foo": 1.0, "bar": "moo"}
        )
        self.assertAlmostEqualGeneric(1.00000001, 1.0, msg="Test message")

    def test_is_testing(self):
        self.assertTrue(is_testing())

    def test_assert_dir_structure(self):
        correct_structure = {
            "extensionless file": None,
            "text file.txt": None,
            "dir": {"another file": None, "empty dir.ext": {},},
        }
        substructure = copy.deepcopy(correct_structure)
        del substructure["dir"]
        wrong_structure = copy.deepcopy(correct_structure)
        wrong_structure["extensionless file"] = {}
        wrong_format = copy.deepcopy(correct_structure)
        wrong_format["dir"]["another_file"] = "foo"

        # Exact structure checks
        ProcessTestCase._assert_dir_structure(
            self.dummy_case, self.temp_dir, correct_structure
        )

        self.assertRaisesRegex(
            AssertionError,
            r"Directory structure mismatch \(exact check\).",
            ProcessTestCase._assert_dir_structure,
            self.dummy_case,
            self.temp_dir,
            substructure,
        )

        self.assertRaisesRegex(
            AssertionError,
            r"Directory structure mismatch \(exact check\).",
            ProcessTestCase._assert_dir_structure,
            self.dummy_case,
            self.temp_dir,
            wrong_structure,
        )

        self.assertRaisesRegex(
            AssertionError,
            "Directory structure specification is incorrect.",
            ProcessTestCase._assert_dir_structure,
            self.dummy_case,
            self.temp_dir,
            wrong_format,
        )

        # Partial structure checks
        ProcessTestCase._assert_dir_structure(
            self.dummy_case, self.temp_dir, correct_structure, False
        )

        ProcessTestCase._assert_dir_structure(
            self.dummy_case, self.temp_dir, substructure, False
        )

        self.assertRaisesRegex(
            AssertionError,
            r"Directory structure mismatch \(partial structure check\).",
            ProcessTestCase._assert_dir_structure,
            self.dummy_case,
            self.temp_dir,
            wrong_structure,
            False,
        )

        self.assertRaisesRegex(
            AssertionError,
            "Directory structure specification is incorrect.",
            ProcessTestCase._assert_dir_structure,
            self.dummy_case,
            self.temp_dir,
            wrong_format,
            False,
        )

    def test_assert_dir_structure_missing(self):
        obj_mock = mock.MagicMock()
        obj_mock.pk = "no id"
        obj_mock.output = {
            "field_name": {"dir": "non-existing directory",},
        }
        obj_mock.process.output_schema = [
            {"name": "field_name", "type": "basic:dir:",},
        ]

        self.assertRaisesRegex(
            AssertionError,
            "Directory .* does not exist.",
            ProcessTestCase.assertDirStructure,
            self.dummy_case,
            obj_mock,
            "field_name",
            "foo.tar.gz",
        )

    def test_assert_dir_missing(self):
        obj_mock = mock.MagicMock()
        obj_mock.pk = "no id"
        obj_mock.output = {"field_name": {"dir": "non-existing directory"}}
        obj_mock.process.output_schema = [{"name": "field_name", "type": "basic:dir:"}]

        self.assertRaisesRegex(
            AssertionError,
            "Directory .* does not exist.",
            ProcessTestCase.assertDir,
            self.dummy_case,
            obj_mock,
            "field_name",
            "foo.tar.gz",
        )

    def test_assert_dir_compressed_missing(self):
        test_output = io.BytesIO()
        correct_output = io.BytesIO()
        open_mock = mock.MagicMock(
            return_value=tarfile.open(fileobj=test_output, mode="w:gz")
        )

        with mock.patch.object(tarfile, "open", open_mock):
            self.assertRaisesRegex(
                AssertionError,
                "Compressed output directory .* missing so it was created.",
                ProcessTestCase._assert_dir,
                self.dummy_case,
                self.temp_dir,
                "",
            )

        with tarfile.open(fileobj=correct_output, mode="w:gz") as f:
            for content in os.listdir(self.temp_dir):
                f.add(os.path.join(self.temp_dir, content), arcname=content)

        self.assertEqual(test_output.getvalue(), correct_output.getvalue())


class TestingFrameworkProcessTestCase(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR])

    @tag_process("test-file-upload")
    def test_url_file(self):
        url = "https://www.genialis.com/data.csv.gz"
        inputs = {
            "src": url,
        }
        datum = self.run_process("test-file-upload", inputs)
        datum.refresh_from_db()

        self.assertFields(datum, "file_temp", url)
        self.assertFields(datum, "file", url)

    @tag_process("test-file-upload")
    def test_do_not_mutate_inputs(self):
        inputs = {
            "src": "example_file.txt",
        }
        original_inputs = inputs.copy()

        self.files_path = TEST_FILES_DIR
        self.run_process("test-file-upload", inputs)

        self.assertEqual(original_inputs, inputs)

    @tag_process("test-file-upload")
    def test_assert_files_exist(self):
        self.files_path = TEST_FILES_DIR
        data = self.run_process("test-file-upload", {"src": "example_file.txt"})
        self.assertFilesExist(data, "files")



""".. Ignore pydocstyle D400.

=================
Permissions utils
=================

.. autofunction:: copy_permissions

"""
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models import Q

from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm, remove_perm
from rest_framework import exceptions


def get_perm_action(perm):
    """Split action from permission of format.

    Input permission can be in the form of '<action>_<object_type>'
    or '<action>'.
    """
    return perm.split("_", 1)[0]


def get_full_perm(perm, obj):
    """Join action with the content type of ``obj``.

    Permission is returned in the format of ``<action>_<object_type>``.
    """
    ctype = ContentType.objects.get_for_model(obj)
    # Camel case class names are converted into a space-separated
    # content types, so spaces have to be removed.
    ctype = str(ctype).replace(" ", "")

    return "{}_{}".format(perm.lower(), ctype)


def get_all_perms(obj):
    """Return a list of all permissions on ``obj``."""
    return list(zip(*obj._meta.permissions))[0]


def change_perm_ctype(perm, dest_obj):
    """Keep permission action and change content type to ``dest_obj``."""
    action = get_perm_action(perm)
    return get_full_perm(action, dest_obj)


def copy_permissions(src_obj, dest_obj):
    """Copy permissions form ``src_obj`` to ``dest_obj``."""

    def _process_permission(codename, user_or_group, dest_obj, relabel):
        """Process single permission."""
        if relabel:
            codename = change_perm_ctype(codename, dest_obj)
            if codename not in dest_all_perms:
                return  # dest object doesn't have matching permission

        assign_perm(codename, user_or_group, dest_obj)

    if src_obj is None:
        return

    src_obj_ctype = ContentType.objects.get_for_model(src_obj)
    dest_obj_ctype = ContentType.objects.get_for_model(dest_obj)
    dest_all_perms = get_all_perms(dest_obj)

    relabel = src_obj_ctype != dest_obj_ctype

    for perm in UserObjectPermission.objects.filter(
        object_pk=src_obj.pk, content_type=src_obj_ctype
    ):
        _process_permission(perm.permission.codename, perm.user, dest_obj, relabel)

    for perm in GroupObjectPermission.objects.filter(
        object_pk=src_obj.pk, content_type=src_obj_ctype
    ):
        _process_permission(perm.permission.codename, perm.group, dest_obj, relabel)


def fetch_user(query):
    """Get user by ``pk``, ``username`` or ``email``.

    Raise error if user can not be determined.
    """
    lookup = Q(username=query) | Q(email=query)
    if query.isdigit():
        lookup = lookup | Q(pk=query)

    user_model = get_user_model()
    users = user_model.objects.filter(lookup)

    if not users:
        raise exceptions.ParseError("Unknown user: {}".format(query))
    elif len(users) >= 2:
        raise exceptions.ParseError("Cannot uniquely determine user: {}".format(query))
    return users[0]


def fetch_group(query):
    """Get group by ``pk`` or ``name``. Raise error if it doesn't exist."""
    group_filter = {"pk": query} if query.isdigit() else {"name": query}

    try:
        return Group.objects.get(**group_filter)
    except Group.DoesNotExist:
        raise exceptions.ParseError("Unknown group: {}".format(query))


def check_owner_permission(payload, allow_user_owner):
    """Raise ``PermissionDenied``if ``owner`` found in ``data``."""
    for entity_type in ["users", "groups"]:
        for perm_type in ["add", "remove"]:
            for perms in payload.get(entity_type, {}).get(perm_type, {}).values():
                if "owner" in perms:
                    if entity_type == "users" and allow_user_owner:
                        continue

                    if entity_type == "groups":
                        raise exceptions.ParseError(
                            "Owner permission cannot be assigned to a group"
                        )

                    raise exceptions.PermissionDenied(
                        "Only owners can grant/revoke owner permission"
                    )


def check_public_permissions(payload):
    """Raise ``PermissionDenied`` if public permissions are too open."""
    allowed_public_permissions = ["view", "add", "download"]
    for perm_type in ["add", "remove"]:
        for perm in payload.get("public", {}).get(perm_type, []):
            if perm not in allowed_public_permissions:
                raise exceptions.PermissionDenied(
                    "Permissions for public users are too open"
                )


def check_user_permissions(payload, user_pk):
    """Raise ``PermissionDenied`` if ``payload`` includes ``user_pk``."""
    for perm_type in ["add", "remove"]:
        user_pks = payload.get("users", {}).get(perm_type, {}).keys()
        if user_pk in user_pks:
            raise exceptions.PermissionDenied("You cannot change your own permissions")


def update_permission(obj, data):
    """Update object permissions."""
    full_permissions = get_all_perms(obj)

    def apply_perm(perm_func, perms, entity):
        """Apply permissions using given ``perm_func``.

        ``perm_func`` is intended to be ``assign_perms`` or
        ``remove_perms`` shortcut function from ``django-guardian``, but
        can be any function that accepts permission codename,
        user/group and object parameters (in this order).

        If given permission does not exist, ``exceptions.ParseError`` is
        raised.

        "ALL" passed as ``perms`` parameter, will call ``perm_function``
        with ``full_permissions`` list.

        :param func perm_func: Permissions function to be applied
        :param list params: list of params to be allpied
        :param entity: user or group to be passed to ``perm_func``
        :type entity: `~django.contrib.auth.models.User` or
            `~django.contrib.auth.models.Group`

        """
        if perms == "ALL":
            perms = full_permissions
        for perm in perms:
            perm_codename = get_full_perm(perm, obj)
            if perm_codename not in full_permissions:
                raise exceptions.ParseError("Unknown permission: {}".format(perm))
            perm_func(perm_codename, entity, obj)

    def set_permissions(entity_type, perm_type):
        """Set object permissions."""
        perm_func = assign_perm if perm_type == "add" else remove_perm
        fetch_fn = fetch_user if entity_type == "users" else fetch_group

        for entity_id in data.get(entity_type, {}).get(perm_type, []):
            entity = fetch_fn(entity_id)
            if entity:
                perms = data[entity_type][perm_type][entity_id]
                apply_perm(perm_func, perms, entity)

    def set_public_permissions(perm_type):
        """Set public permissions."""
        perm_func = assign_perm if perm_type == "add" else remove_perm
        user = AnonymousUser()
        perms = data.get("public", {}).get(perm_type, [])
        apply_perm(perm_func, perms, user)

    with transaction.atomic():
        set_permissions("users", "add")
        set_permissions("users", "remove")
        set_permissions("groups", "add")
        set_permissions("groups", "remove")
        set_public_permissions("add")
        set_public_permissions("remove")


def assign_contributor_permissions(obj, contributor=None):
    """Assign all permissions to object's contributor."""
    for permission in get_all_perms(obj):
        assign_perm(permission, contributor if contributor else obj.contributor, obj)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-26 04:08
from __future__ import unicode_literals

from django.db import migrations, models


def set_data_relation(apps, schema_editor):
    Storage = apps.get_model("flow", "Storage")

    for storage in Storage.objects.iterator():
        storage.data.add(storage.data_migration_temporary)


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0028_add_data_location"),
    ]

    operations = [
        migrations.RenameField(
            model_name="storage", old_name="data", new_name="data_migration_temporary",
        ),
        migrations.AddField(
            model_name="storage",
            name="data",
            field=models.ManyToManyField(related_name="storages", to="flow.Data"),
        ),
        migrations.RunPython(set_data_relation),
        migrations.RemoveField(model_name="storage", name="data_migration_temporary",),
    ]



# pylint: disable=missing-docstring
import os

from resolwe.test import ProcessTestCase, tag_process, with_docker_executor


class FilesProcessTestCase(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self.files_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "files")
        )

    @with_docker_executor
    @tag_process("upload-file")
    def test_upload_file(self):
        upload_file = self.run_process("upload-file", {"src": "file binary"})
        self.assertFile(upload_file, "file", "file binary")

    @with_docker_executor
    @tag_process("upload-file")
    def test_upload_file_nested_path(self):
        """Test that nested files from files directory are processed correctly."""
        upload_file = self.run_process(
            "upload-file", {"src": "upload_file/input_file.txt"}
        )
        self.assertFile(upload_file, "file", "upload_file/input_file.txt")

    @with_docker_executor
    @tag_process("upload-image-file")
    def test_upload_image(self):
        upload_image = self.run_process("upload-image-file", {"src": "file image.png"})
        self.assertFile(upload_image, "file", "file image.png")

    @with_docker_executor
    @tag_process("upload-tab-file")
    def test_upload_tabular_tab(self):
        upload_tab = self.run_process("upload-tab-file", {"src": "file tab.tab.gz"})
        self.assertFile(
            upload_tab, "file", "tab_file_tabular.tab.gz", compression="gzip"
        )
        self.assertFile(upload_tab, "src_file", "file tab.tab.gz", compression="gzip")

    @with_docker_executor
    @tag_process("upload-tab-file")
    def test_upload_tabular_csv(self):
        upload_csv = self.run_process("upload-tab-file", {"src": "file tab.csv.gz"})
        self.assertFile(
            upload_csv, "file", "csv_file_tabular.tab.gz", compression="gzip"
        )
        self.assertFile(upload_csv, "src_file", "file tab.csv.gz", compression="gzip")

    @with_docker_executor
    @tag_process("upload-tab-file")
    def test_upload_tabular_xls(self):
        upload_xls = self.run_process("upload-tab-file", {"src": "file tab.xls"})
        self.assertFile(
            upload_xls, "file", "xls_file_tabular.tab.gz", compression="gzip"
        )
        self.assertFile(upload_xls, "src_file", "file tab.xls")

    @with_docker_executor
    @tag_process("upload-tab-file")
    def test_upload_tabular_xlsx(self):
        upload_xlsx = self.run_process("upload-tab-file", {"src": "file tab.1.xlsx"})
        self.assertFile(
            upload_xlsx, "file", "xlsx_file_tabular.tab.gz", compression="gzip"
        )
        self.assertFile(upload_xlsx, "src_file", "file tab.1.xlsx")



""".. Ignore pydocstyle D400.

============
Test Helpers
============

.. automodule:: resolwe.test_helpers.management.commands.list_process_tags
    :members:

.. automodule:: resolwe.test_helpers.management.commands.show_profile
    :members:

"""



""".. Ignore pydocstyle D400.

============
Flow Filters
============

"""
import django_filters as filters

from rest_framework.exceptions import ParseError

from .models import Collection, Data, DescriptorSchema, Entity, Process, Relation

NUMBER_LOOKUPS = [
    "exact",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
    "isnull",
]
TEXT_LOOKUPS = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "in",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "isnull",
]
DATE_LOOKUPS = [
    "exact",
    "gt",
    "gte",
    "lt",
    "lte",
    "year",
    "year__gt",
    "year__gte",
    "year__lt",
    "year__lte",
    "month",
    "month__gt",
    "month__gte",
    "month__lt",
    "month__lte",
    "day",
    "day__gt",
    "day__gte",
    "day__lt",
    "day__lte",
    "isnull",
]
DATETIME_LOOKUPS = DATE_LOOKUPS + [
    "date",
    "time",
    "hour",
    "hour__gt",
    "hour__gte",
    "hour__lt",
    "hour__lte",
    "minute",
    "minute__gt",
    "minute__gte",
    "minute__lt",
    "minute__lte",
    "second",
    "second__gt",
    "second__gte",
    "second__lt",
    "second__lte",
]


class CheckQueryParamsMixin:
    """Custom query params validation."""

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return (
            "fields",
            "format",
            "limit",
            "offset",
            "ordering",
        )

    def validate_query_params(self):
        """Ensure no unsupported query params were used."""
        allowed_params = set(self.get_filters().keys())
        allowed_params.update(self.get_always_allowed_arguments())

        unallowed = set(self.request.query_params.keys()) - allowed_params

        if unallowed:
            msg = "Unsupported parameter(s): {}. Please use a combination of: {}.".format(
                ", ".join(unallowed), ", ".join(allowed_params),
            )
            self.form.add_error(field=None, error=ParseError(msg))

    def is_valid(self):
        """Validate filterset."""
        self.validate_query_params()
        return super().is_valid()


class BaseResolweFilter(CheckQueryParamsMixin, filters.FilterSet):
    """Base filter for Resolwe's endpoints."""

    class Meta:
        """Filter configuration."""

        fields = {
            "id": NUMBER_LOOKUPS[:],
            "slug": TEXT_LOOKUPS[:],
            "name": TEXT_LOOKUPS[:],
            "contributor": ["exact", "in"],
            "created": DATETIME_LOOKUPS[:],
            "modified": DATETIME_LOOKUPS[:],
        }


class DescriptorSchemaFilter(BaseResolweFilter):
    """Filter the DescriptorSchema endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = DescriptorSchema


class CollectionFilter(BaseResolweFilter):
    """Filter the Collection endpoint."""

    data = filters.ModelChoiceFilter(queryset=Data.objects.all())
    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Collection
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{"description": TEXT_LOOKUPS[:], "descriptor_schema": ["exact"],},
        }


class TagsFilter(filters.filters.BaseCSVFilter, filters.CharFilter):
    """Filter for tags."""

    def __init__(self, *args, **kwargs):
        """Construct tags filter."""
        kwargs.setdefault("lookup_expr", "contains")
        super().__init__(*args, **kwargs)


class EntityFilter(CollectionFilter):
    """Filter the Entity endpoint."""

    collection = filters.ModelChoiceFilter(
        field_name="collection", queryset=Collection.objects.all()
    )
    tags = TagsFilter()

    class Meta(CollectionFilter.Meta):
        """Filter configuration."""

        model = Entity


class ProcessFilter(BaseResolweFilter):
    """Filter the Process endpoint."""

    category = filters.CharFilter(field_name="category", lookup_expr="startswith")
    type = filters.CharFilter(field_name="type", lookup_expr="startswith")
    is_active = filters.rest_framework.filters.BooleanFilter(field_name="is_active")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Process
        fields = {**BaseResolweFilter.Meta.fields, **{"scheduling_class": ["exact"],}}


class CharInFilter(filters.BaseInFilter, filters.CharFilter):
    """Helper class for creation of CharFilter with "in" lookup."""


class DataFilter(BaseResolweFilter):
    """Filter the Data endpoint."""

    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    collection__slug = filters.CharFilter(
        field_name="collection__slug", lookup_expr="exact"
    )

    entity = filters.ModelChoiceFilter(queryset=Entity.objects.all())

    type = filters.CharFilter(field_name="process__type", lookup_expr="startswith")
    status = filters.CharFilter(lookup_expr="iexact")
    status__in = CharInFilter(field_name="status", lookup_expr="in")

    tags = TagsFilter()

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Data
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "process": ["exact"],
                "process__slug": ["exact"],
                "finished": DATETIME_LOOKUPS[:],
                "started": DATETIME_LOOKUPS[:],
            },
        }


class RelationFilter(BaseResolweFilter):
    """Filter the Relation endpoint."""

    category = filters.CharFilter(lookup_expr="iexact")
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(field_name="type__name")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Relation
        fields = BaseResolweFilter.Meta.fields

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return super().get_always_allowed_arguments() + ("entity", "label", "position",)



# pylint: disable=missing-docstring,too-many-lines
import io
import json
import os
import shutil
from datetime import timedelta
from unittest.mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.timezone import now

from guardian.shortcuts import assign_perm, get_perms, remove_perm
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models import (
    Collection,
    DataMigrationHistory,
    DescriptorSchema,
    Entity,
    Process,
    Storage,
)
from resolwe.flow.models.data import (
    Data,
    DataDependency,
    DataLocation,
    hydrate_size,
    render_template,
)
from resolwe.flow.models.utils import hydrate_input_references
from resolwe.flow.views import DataViewSet
from resolwe.test import TestCase, TransactionTestCase

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2


class DataModelNameTest(TransactionTestCase):
    def test_name(self):
        process = Process.objects.create(
            slug="test-first",
            type="data:test:first:",
            contributor=self.contributor,
            data_name="Process based data name",
            output_schema=[
                {"name": "stat", "type": "basic:string:", "required": False,}
            ],
            run={"language": "bash", "program": 'echo {"stat": "42"}'},
        )

        data = Data.objects.create(contributor=self.contributor, process=process)
        data.refresh_from_db()

        self.assertEqual(data.name, "Process based data name")
        self.assertFalse(data.named_by_user)

        data.name = "Name changed by user"
        data.save()
        data.refresh_from_db()

        self.assertEqual(data.name, "Name changed by user")
        self.assertTrue(data.named_by_user)

        data = Data.objects.create(
            name="Explicit data name", contributor=self.contributor, process=process
        )
        data.refresh_from_db()

        self.assertEqual(data.name, "Explicit data name")
        self.assertTrue(data.named_by_user)

        process = Process.objects.create(
            slug="test-second",
            type="test:second",
            contributor=self.contributor,
            requirements={"expression-engine": "jinja"},
            data_name="Process based data name, value: {{src.stat}}",
            input_schema=[
                {"name": "src", "type": "data:test:first:", "required": False,}
            ],
        )

        with transaction.atomic():
            second = Data.objects.create(
                contributor=self.contributor, process=process, input={"src": data.id}
            )

            data.output = {"stat": "42"}
            data.status = "OK"
            data.save()

            self.assertEqual(second.name, "Process based data name, value: ")
            self.assertFalse(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, "Process based data name, value: 42")
        self.assertFalse(second.named_by_user)

        with transaction.atomic():
            data.output = {}
            data.status = "RE"
            data.save()

            second = Data.objects.create(
                contributor=self.contributor, process=process, input={"src": data.id}
            )

            second.name = "User' data name"
            second.save()

            data.output = {"stat": "42"}
            data.status = "OK"
            data.save()

            self.assertEqual(second.name, "User' data name")
            self.assertTrue(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, "User' data name")
        self.assertTrue(second.named_by_user)


class DataModelTest(TestCase):
    def test_delete_chunked(self):
        process = Process.objects.create(contributor=self.contributor)
        for _ in range(1000):
            Data.objects.create(contributor=self.contributor, process=process)

        Data.objects.all().delete_chunked(chunk_size=50)
        self.assertFalse(Data.objects.exists())

    def test_trim_name(self):
        process = Process.objects.create(
            contributor=self.contributor, data_name="test" * 50
        )
        data = Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(len(data.name), 100)
        self.assertEqual(data.name[-3:], "...")

    def test_hydrate_file_size(self):
        proc = Process.objects.create(
            name="Test process",
            contributor=self.contributor,
            output_schema=[{"name": "output_file", "type": "basic:file:"}],
        )

        data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=proc,
        )

        data.output = {"output_file": {"file": "output.txt"}}

        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        dir_path = data.location.get_path()
        os.makedirs(dir_path)

        with self.assertRaises(ValidationError):
            data.save()

        file_path = os.path.join(dir_path, "output.txt")
        with open(file_path, "w") as fn:
            fn.write("foo bar")

        data.save()

        self.assertEqual(data.output["output_file"]["size"], 7)

    def test_dependencies_single(self):
        process = Process.objects.create(
            slug="test-dependencies",
            type="data:test:dependencies:",
            contributor=self.contributor,
            input_schema=[
                {"name": "src", "type": "data:test:dependencies:", "required": False,}
            ],
        )

        first = Data.objects.create(
            name="First data", contributor=self.contributor, process=process
        )

        second = Data.objects.create(
            name="Second data",
            contributor=self.contributor,
            process=process,
            input={"src": first.id},
        )

        third = Data.objects.create(
            name="Third data",
            contributor=self.contributor,
            process=process,
            input={"src": first.id},
        )

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 0)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 1)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO}
        )
        self.assertEqual(
            {d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO}
        )
        self.assertEqual(
            {d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO}
        )

    def test_dependencies_list(self):
        process = Process.objects.create(
            slug="test-dependencies-list",
            type="data:test:dependencies:list:",
            contributor=self.contributor,
            input_schema=[
                {
                    "name": "src",
                    "type": "list:data:test:dependencies:list:",
                    "required": False,
                }
            ],
        )

        first = Data.objects.create(
            name="First data", contributor=self.contributor, process=process
        )

        second = Data.objects.create(
            name="Second data",
            contributor=self.contributor,
            process=process,
            input={"src": [first.id]},
        )

        third = Data.objects.create(
            name="Third data",
            contributor=self.contributor,
            process=process,
            input={"src": [first.id, second.id]},
        )

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 1)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 2)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())
        self.assertIn(third, second.children.all())

        # Check correct dependency type is created.
        self.assertEqual(
            {d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO}
        )
        self.assertEqual(
            {d.kind for d in second.children_dependency.all()}, {DataDependency.KIND_IO}
        )
        self.assertEqual(
            {d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO}
        )
        self.assertEqual(
            {d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO}
        )


class EntityModelTest(TestCase):
    def setUp(self):
        super().setUp()

        DescriptorSchema.objects.create(
            name="Sample", slug="sample", contributor=self.contributor
        )
        self.process = Process.objects.create(
            name="Test process",
            type="data:test:",
            contributor=self.contributor,
            entity_type="sample",
            entity_descriptor_schema="sample",
        )
        # Entity is created automatically when Data object is created
        self.data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=self.process
        )

    def test_delete_data(self):
        # Create another Data object and add it to the same Entity.
        data_2 = Data.objects.create(
            name="Test data 2", contributor=self.contributor, process=self.process
        )
        data_2.entity.delete()
        data_2.entity = self.data.entity
        data_2.save()
        data_2.delete()

        self.assertEqual(Entity.objects.count(), 1)

    def test_new_sample(self):
        data = Data.objects.create(
            name="Test data",
            contributor=self.contributor,
            process=self.process,
            tags=["foo", "bar"],
        )
        entity = Entity.objects.last()
        self.assertTrue(entity.data.filter(pk=data.pk).exists())

        # Make sure tags are copied.
        self.assertEqual(entity.tags, data.tags)

    def test_entity_inheritance(self):
        # Prepare Data objects with entities.
        process_entity = self.process
        data_1 = self.data
        data_2 = Data.objects.create(
            contributor=self.contributor, process=process_entity
        )
        # Prepare Data objects without entities.
        process_no_entity = Process.objects.create(
            type="data:test:", contributor=self.contributor
        )
        data_3 = Data.objects.create(
            contributor=self.contributor, process=process_no_entity
        )
        # Prepare test process.
        test_process = Process.objects.create(
            contributor=self.contributor,
            entity_type="sample",
            entity_descriptor_schema="sample",
            input_schema=[
                {"name": "data_list", "type": "list:data:test:", "required": False},
                {"name": "data", "type": "data:test:", "required": False},
            ],
        )

        # Single Entity - Data object should be added to it.
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data_list": [data_1.pk, data_3.pk],},
        )
        self.assertEqual(data.entity, data_1.entity)

        # Multiple Entities - Data object should be added to none of them.
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data_list": [data_1.pk, data_2.pk],},
        )
        self.assertEqual(data.entity, None)

        # Multiple Entities with entity_always_create = True - Data object should be added to a whole new entity.
        test_process.entity_always_create = True
        test_process.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data": data_1.pk, "data_list": [data_2.pk, data_3.pk],},
        )
        self.assertEqual(data.entity, Entity.objects.last())
        test_process.entity_always_create = False
        test_process.save()

        # Multiple Entities with entity_input defined - Data object should be added to it.
        test_process.entity_input = "data"
        test_process.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data": data_1.pk, "data_list": [data_2.pk],},
        )
        self.assertEqual(data.entity, data_1.entity)
        test_process.entity_input = "data_list"
        test_process.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data": data_1.pk, "data_list": [data_2.pk],},
        )
        self.assertEqual(data.entity, data_2.entity)
        test_process.entity_input = None
        test_process.save()

        # Entities of different types - Data object should be added to the right one.
        entity_2 = data_2.entity
        entity_2.type = "something_else"
        entity_2.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={"data_list": [data_1.pk, data_2.pk],},
        )
        self.assertEqual(data.entity, data_1.entity)
        entity_2.type = "sample"
        entity_2.save()


class GetOrCreateTestCase(APITestCase):
    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        self.user = user_model.objects.create(username="test_user", password="test_pwd")

        self.process = Process.objects.create(
            name="Temporary process",
            contributor=self.user,
            slug="tmp-process",
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[
                {"name": "some_value", "type": "basic:integer:", "default": 42}
            ],
        )
        assign_perm("view_process", self.user, self.process)

        process_2 = Process.objects.create(
            name="Another process",
            contributor=self.user,
            slug="another-process",
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[{"name": "some_value", "type": "basic:integer:"}],
        )
        assign_perm("view_process", self.user, process_2)

        self.data = Data.objects.create(
            name="Temporary data",
            contributor=self.user,
            process=self.process,
            input={"some_value": 42},
        )
        assign_perm("view_data", self.user, self.data)

        self.get_or_create_view = DataViewSet.as_view({"post": "get_or_create"})
        self.factory = APIRequestFactory()

    def tearDown(self):
        for data in Data.objects.all():
            if data.location:
                data_dir = data.location.get_path()
                shutil.rmtree(data_dir, ignore_errors=True)

        super().tearDown()

    def test_get_same(self):
        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 42},
                "process": {"slug": "tmp-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], self.data.pk)

    def test_use_defaults(self):
        request = self.factory.post(
            "",
            {"name": "Data", "input": {}, "process": {"slug": "tmp-process"}},
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], self.data.pk)

    def test_missing_permission(self):
        remove_perm("view_data", self.user, self.data)

        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 42},
                "process": {"slug": "tmp-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data["id"], self.data.pk)

    def test_different_input(self):
        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 43},
                "process": {"slug": "tmp-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data["id"], self.data.pk)

    def test_different_process(self):
        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 43},
                "process": {"slug": "another-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data["id"], self.data.pk)

    def test_different_process_version(self):
        self.process.version = "2.0.0"
        self.process.save()

        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 42},
                "process": {"slug": "tmp-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data["id"], self.data.pk)

    def test_raw_process(self):
        self.process.persistence = Process.PERSISTENCE_RAW
        self.process.save()

        request = self.factory.post(
            "",
            {
                "name": "Data",
                "input": {"some_value": 42},
                "process": {"slug": "tmp-process"},
            },
            format="json",
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data["id"], self.data.pk)


class DuplicateTestCase(TestCase):
    def test_data_duplicate(self):
        process1 = Process.objects.create(
            contributor=self.user, type="data:test:first:",
        )
        process2 = Process.objects.create(
            contributor=self.user,
            input_schema=[{"name": "data_field", "type": "data:test:first:"},],
            output_schema=[{"name": "json_field", "type": "basic:json:"},],
        )

        input_data = Data.objects.create(
            name="Data 1", contributor=self.user, process=process1,
        )
        input_data.status = Data.STATUS_DONE
        input_data.save()

        data2 = Data.objects.create(
            name="Data 2",
            contributor=self.user,
            process=process2,
            started=now(),
            input={"data_field": input_data.id},
        )
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data2)
        data2.output = {"json_field": {"foo": "bar"}}
        data2.status = Data.STATUS_DONE
        data2.save()
        data2.migration_history.create(migration="migration_1")
        data2.migration_history.create(migration="migration_2")

        # Duplicate.
        duplicates = Data.objects.filter(id=data2.id).duplicate(self.contributor)
        self.assertEqual(len(duplicates), 1)
        duplicate = duplicates[0]
        self.assertTrue(duplicate.is_duplicate())

        # Convert original and duplicated to dict.
        data2_dict = Data.objects.filter(id=data2.id).values()[0]
        duplicate_dict = Data.objects.filter(id=duplicate.id).values()[0]

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = (
            "id",
            "slug",
            "contributor_id",
            "name",
            "duplicated",
            "modified",
        )
        for model_dict in (data2_dict, duplicate_dict):
            for field in fields_to_differ:
                model_dict.pop(field)

        self.assertDictEqual(data2_dict, duplicate_dict)

        # Assert location.
        self.assertEqual(data2.location.id, duplicate.location.id)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, "copy-of-data-2")
        self.assertEqual(duplicate.name, "Copy of Data 2")
        self.assertEqual(duplicate.contributor.username, "contributor")
        self.assertAlmostEqual(duplicate.duplicated, now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, now(), delta=timedelta(seconds=3))

        # Assert dependencies.
        self.assertEqual(DataDependency.objects.count(), 2)
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO, parent=input_data, child=data2
            ).exists()
        )
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO, parent=input_data, child=duplicate
            ).exists()
        )

        # Assert storage
        self.assertEqual(Storage.objects.count(), 1)
        self.assertEqual(data2.storages.first(), Storage.objects.first())
        self.assertEqual(duplicate.storages.first(), Storage.objects.first())

        # Assert migration history
        self.assertEqual(DataMigrationHistory.objects.count(), 4)
        self.assertEqual(
            data2.migration_history.earliest("created").migration, "migration_1"
        )
        self.assertEqual(
            data2.migration_history.latest("created").migration, "migration_2"
        )
        self.assertEqual(
            duplicate.migration_history.earliest("created").migration, "migration_1"
        )
        self.assertEqual(
            duplicate.migration_history.latest("created").migration, "migration_2"
        )

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 4)

    def test_data_duplicate_duplicate(self):
        process1 = Process.objects.create(
            contributor=self.user, type="data:test:first:",
        )
        process2 = Process.objects.create(
            contributor=self.user,
            input_schema=[{"name": "data_field", "type": "data:test:first:"},],
            output_schema=[{"name": "json_field", "type": "basic:json:"},],
        )

        input_data = Data.objects.create(contributor=self.user, process=process1)

        data = Data.objects.create(
            contributor=self.user, process=process2, input={"data_field": input_data.id}
        )
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)
        data.output = {"json_field": {"foo": "bar"}}
        data.status = Data.STATUS_DONE
        data.save()
        data.migration_history.create(migration="migration_1")

        # Duplicate.
        duplicate = data.duplicate(self.contributor)
        duplicate_of_duplicate = duplicate.duplicate()

        self.assertEqual(
            duplicate.contributor.id, duplicate_of_duplicate.contributor.id
        )

        # Assert location
        self.assertEqual(duplicate.location.id, duplicate_of_duplicate.location.id)

        # Assert dependencies.
        self.assertEqual(DataDependency.objects.count(), 3)
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO,
                parent=input_data,
                child=duplicate_of_duplicate,
            ).exists()
        )

        # Assert storage
        self.assertEqual(
            duplicate_of_duplicate.storages.first().id, Storage.objects.first().id
        )

        # Assert migration history
        self.assertEqual(DataMigrationHistory.objects.count(), 3)
        self.assertEqual(
            duplicate_of_duplicate.migration_history.latest("created").migration,
            "migration_1",
        )

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate_of_duplicate)), 4)

    def test_input_rewiring(self):
        process1 = Process.objects.create(
            contributor=self.user, type="data:test:first:",
        )
        process2 = Process.objects.create(
            contributor=self.user,
            type="data:test:second:",
            input_schema=[
                {"name": "data_field1", "type": "data:test:first:"},
                {"name": "data_field2", "type": "data:test:first:"},
                {"name": "data_list_field", "type": "list:data:test:first:"},
            ],
        )

        data1 = Data.objects.create(
            contributor=self.user, process=process1, status=Data.STATUS_DONE,
        )

        should_not_be_rewritten = data1.duplicate()

        data2 = Data.objects.create(
            contributor=self.user,
            process=process2,
            input={
                "data_field1": data1.id,
                "data_field2": should_not_be_rewritten.id,
                "data_list_field": [data1.id, should_not_be_rewritten.id],
            },
            status=Data.STATUS_DONE,
        )

        # Duplicate.
        duplicates = (
            Data.objects.filter(id__in=[data1.id, data2.id])
            .order_by("id")
            .duplicate(self.contributor)
        )
        duplicate1, duplicate2 = duplicates

        self.assertEqual(duplicate2.input["data_field1"], duplicate1.id)
        self.assertEqual(duplicate2.input["data_field2"], should_not_be_rewritten.id)
        self.assertEqual(
            duplicate2.input["data_list_field"],
            [duplicate1.id, should_not_be_rewritten.id],
        )

    def test_data_status_not_done(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(
            contributor=self.user, process=process, status=Data.STATUS_WAITING
        )
        data2 = Data.objects.create(
            contributor=self.user, process=process, status=Data.STATUS_DONE
        )

        with self.assertRaisesMessage(
            ValidationError,
            "Data object must have done or error status to be duplicated",
        ):
            Data.objects.filter(id__in=[data.id, data2.id]).duplicate(self.contributor)

    def test_data_long_name(self):
        process = Process.objects.create(contributor=self.user)

        long_name = "a" * (Data._meta.get_field("name").max_length - 1)
        data = Data.objects.create(
            name=long_name,
            contributor=self.user,
            process=process,
            status=Data.STATUS_DONE,
        )

        duplicate = Data.objects.filter(id=data.id).duplicate(self.contributor)[0]

        self.assertTrue(duplicate.name.startswith("Copy of "))
        self.assertTrue(duplicate.name.endswith("..."))
        self.assertEqual(len(duplicate.name), Data._meta.get_field("name").max_length)

    def test_entity_duplicate(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(
            name="Data 1",
            contributor=self.user,
            process=process,
            status=Data.STATUS_DONE,
        )
        assign_perm("view_data", self.contributor, data)
        data2 = Data.objects.create(
            name="Data 2",
            contributor=self.user,
            process=process,
            status=Data.STATUS_DONE,
        )
        assign_perm("view_data", self.contributor, data2)
        entity = Entity.objects.create(name="Entity", contributor=self.user)
        entity.data.add(data, data2)

        # Add to collection.
        collection = Collection.objects.create(contributor=self.user)
        collection.entity_set.add(entity)
        data.collection = collection
        data.save()
        data2.collection = collection
        data2.save()

        # Duplicate.
        entities = Entity.objects.filter(id=entity.id).duplicate(self.contributor)
        self.assertEqual(len(entities), 1)
        duplicate = entities[0]

        entity_dict = Entity.objects.filter(id=entity.id).values()[0]
        duplicate_dict = Entity.objects.filter(id=duplicate.id).values()[0]

        data1_dict = Data.objects.filter(id=data.id).values()[0]
        data1_duplicate = duplicate.data.get(name="Copy of Data 1")
        data1_duplicate_dict = Data.objects.filter(id=data1_duplicate.id).values()[0]

        data2_dict = Data.objects.filter(id=data2.id).values()[0]
        data2_duplicate = duplicate.data.get(name="Copy of Data 2")
        data2_duplicate_dict = Data.objects.filter(id=data2_duplicate.id).values()[0]

        self.assertTrue(duplicate.is_duplicate())
        self.assertTrue(data1_duplicate.is_duplicate())

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = (
            "id",
            "slug",
            "contributor_id",
            "name",
            "duplicated",
            "modified",
            "collection_id",
            "entity_id",
        )
        for model_dict in (
            entity_dict,
            duplicate_dict,
            data1_dict,
            data1_duplicate_dict,
            data2_dict,
            data2_duplicate_dict,
        ):
            for field in fields_to_differ:
                model_dict.pop(field, "")

        self.assertDictEqual(entity_dict, duplicate_dict)
        self.assertDictEqual(data1_dict, data1_duplicate_dict)
        self.assertDictEqual(data2_dict, data2_duplicate_dict)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, "copy-of-entity")
        self.assertEqual(duplicate.name, "Copy of Entity")
        self.assertEqual(duplicate.contributor.username, "contributor")
        self.assertAlmostEqual(duplicate.duplicated, now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, now(), delta=timedelta(seconds=3))
        self.assertEqual(duplicate.collection, None)

        # Assert collection is not altered.
        self.assertEqual(Collection.objects.count(), 1)
        self.assertEqual(collection.entity_set.count(), 1)
        self.assertEqual(collection.entity_set.first().name, "Entity")
        self.assertEqual(collection.data.count(), 2)
        collection_data = collection.data.all().order_by("name")
        self.assertEqual(collection_data[0].name, "Data 1")
        self.assertEqual(collection_data[1].name, "Data 2")

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 4)

    def test_entity_duplicate_inherit(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(
            contributor=self.user, process=process, status=Data.STATUS_DONE
        )
        assign_perm("view_data", self.user, data)
        entity = Entity.objects.create(contributor=self.user)
        entity.data.add(data)

        # Add to collection.
        collection = Collection.objects.create(contributor=self.user)
        assign_perm("view_collection", self.user, collection)
        assign_perm("edit_collection", self.user, collection)
        assign_perm("edit_collection", self.contributor, collection)
        collection.entity_set.add(entity)
        collection.data.add(data)

        # Duplicate.
        duplicated_entity1 = Entity.objects.filter(id=entity.id).duplicate(
            self.user, inherit_collection=False
        )[0]

        self.assertEqual(collection.entity_set.count(), 1)
        self.assertEqual(collection.entity_set.first().id, entity.id)
        self.assertEqual(collection.data.count(), 1)
        self.assertEqual(collection.data.first().id, data.id)

        # Assert permissions.
        self.assertEqual(duplicated_entity1.data.count(), 1)
        self.assertCountEqual(
            get_perms(self.user, collection), ["view_collection", "edit_collection"]
        )
        self.assertEqual(len(get_perms(self.user, duplicated_entity1)), 4)
        self.assertEqual(len(get_perms(self.user, duplicated_entity1.data.first())), 4)
        self.assertListEqual(
            get_perms(self.contributor, collection), ["edit_collection"]
        )
        self.assertListEqual(get_perms(self.contributor, duplicated_entity1), [])
        self.assertListEqual(
            get_perms(self.contributor, duplicated_entity1.data.first()), []
        )

        duplicated_entity2 = Entity.objects.filter(id=entity.id).duplicate(
            self.user, inherit_collection=True
        )[0]

        self.assertEqual(collection.entity_set.count(), 2)
        self.assertEqual(collection.entity_set.first().id, entity.id)
        self.assertEqual(collection.entity_set.last().id, duplicated_entity2.id)
        self.assertEqual(collection.data.count(), 2)
        self.assertEqual(collection.data.first().id, data.id)
        self.assertEqual(collection.data.last().id, duplicated_entity2.data.first().id)

        # Assert permissions.
        self.assertEqual(duplicated_entity2.data.count(), 1)
        self.assertCountEqual(
            get_perms(self.user, collection), ["view_collection", "edit_collection"]
        )
        self.assertEqual(len(get_perms(self.user, duplicated_entity2)), 4)
        self.assertEqual(len(get_perms(self.user, duplicated_entity2.data.first())), 4)
        self.assertListEqual(
            get_perms(self.contributor, collection), ["edit_collection"]
        )
        self.assertListEqual(
            get_perms(self.contributor, duplicated_entity2), ["edit_entity"]
        )
        self.assertListEqual(
            get_perms(self.contributor, duplicated_entity2.data.first()), ["edit_data"]
        )

    def test_collection_duplicate(self):
        entity = Entity.objects.create(name="Entity", contributor=self.user)
        assign_perm("view_entity", self.contributor, entity)

        # Add to collection.
        collection = Collection.objects.create(name="Collection", contributor=self.user)
        collection.entity_set.add(entity)

        # Duplicate.
        collections = Collection.objects.filter(id=collection.id).duplicate(
            self.contributor
        )
        self.assertEqual(len(collections), 1)
        duplicate = collections[0]

        collection_dict = Collection.objects.filter(id=collection.id).values()[0]
        duplicate_dict = Collection.objects.filter(id=duplicate.id).values()[0]

        entity_dict = Entity.objects.filter(id=entity.id).values()[0]
        entity_duplicate = duplicate.entity_set.first()
        entity_duplicate_dict = Entity.objects.filter(id=entity_duplicate.id).values()[
            0
        ]

        self.assertTrue(duplicate.is_duplicate())
        self.assertTrue(entity_duplicate.is_duplicate())

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = (
            "id",
            "slug",
            "contributor_id",
            "name",
            "duplicated",
            "modified",
            "collection_id",
        )
        for model_dict in (
            collection_dict,
            duplicate_dict,
            entity_dict,
            entity_duplicate_dict,
        ):
            for field in fields_to_differ:
                model_dict.pop(field, "")

        self.assertDictEqual(collection_dict, duplicate_dict)
        self.assertDictEqual(entity_dict, entity_duplicate_dict)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, "copy-of-collection")
        self.assertEqual(duplicate.name, "Copy of Collection")
        self.assertEqual(duplicate.contributor.username, "contributor")
        self.assertAlmostEqual(duplicate.duplicated, now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, now(), delta=timedelta(seconds=3))

        self.assertEqual(entity_duplicate.slug, "copy-of-entity")
        self.assertEqual(entity_duplicate.name, "Copy of Entity")
        self.assertEqual(entity_duplicate.contributor.username, "contributor")
        self.assertAlmostEqual(
            entity_duplicate.duplicated, now(), delta=timedelta(seconds=3)
        )
        self.assertAlmostEqual(
            entity_duplicate.modified, now(), delta=timedelta(seconds=3)
        )

        # Assert duplicated entity and data objects are in collection.
        self.assertEqual(duplicate.entity_set.count(), 1)
        self.assertEqual(duplicate.entity_set.first().name, "Copy of Entity")

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 4)


class ProcessModelTest(TestCase):
    def test_process_isactive(self):
        process = Process.objects.create(contributor=self.contributor)
        # is_active is true by default
        self.assertIs(process.is_active, True)
        process.is_active = False
        process.save()
        process_fetched = Process.objects.get(pk=process.pk)
        # the is_active flag is saved
        self.assertFalse(process_fetched.is_active)


@patch("resolwe.flow.models.utils.os")
class HydrateFileSizeUnitTest(TestCase):
    def create_data(self, os_mock, contributor, process):
        # Mock isfile and getsize for data creation to pass.
        # Make sure to set these two values to desired
        # values after calling this method.
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 0

        data = Data.objects.create(contributor=contributor, process=process)
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        data.output = {"test_file": {"file": "test_file.tmp"}}
        data.save()

        return data

    def setUp(self):
        super().setUp()

        self.process = Process.objects.create(
            contributor=self.contributor,
            output_schema=[
                {"name": "test_file", "type": "basic:file:", "required": False},
                {"name": "file_list", "type": "list:basic:file:", "required": False},
            ],
        )

    def test_done_data(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(data)
        self.assertEqual(data.output["test_file"]["size"], 42000)

    def test_data_with_refs(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [42000, 8000, 42]
        data.output = {
            "test_file": {
                "file": "test_file.tmp",
                "refs": ["ref_file1.tmp", "ref_file2.tmp"],
            }
        }
        hydrate_size(data)
        self.assertEqual(data.output["test_file"]["size"], 42000)
        self.assertEqual(data.output["test_file"]["total_size"], 50042)
        self.assertEqual(data.size, 50042)

    def test_list(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [34, 42000, 42]
        data.output = {
            "file_list": [
                {"file": "test_01.tmp",},
                {"file": "test_02.tmp", "refs": ["ref_file1.tmp"]},
            ]
        }
        hydrate_size(data)
        self.assertEqual(data.output["file_list"][0]["size"], 34)
        self.assertEqual(data.output["file_list"][0]["total_size"], 34)
        self.assertEqual(data.output["file_list"][1]["size"], 42000)
        self.assertEqual(data.output["file_list"][1]["total_size"], 42042)
        self.assertEqual(data.size, 34 + 42042)

    def test_change_size(self, os_mock):
        """Size is not changed after object is done."""
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(data)
        self.assertEqual(data.output["test_file"]["size"], 42000)

        os_mock.path.getsize.return_value = 43000
        hydrate_size(data)
        self.assertEqual(data.output["test_file"]["size"], 43000)

        data.status = Data.STATUS_DONE
        os_mock.path.getsize.return_value = 44000
        hydrate_size(data)
        self.assertEqual(data.output["test_file"]["size"], 43000)

    def test_missing_file(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = False
        with self.assertRaises(ValidationError):
            hydrate_size(data)


class StorageModelTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.proc = Process.objects.create(
            name="Test process",
            contributor=self.contributor,
            output_schema=[{"name": "json_field", "type": "basic:json:"},],
        )

    def test_save_storage(self):
        """`basic:json:` fields are stored in Storage"""
        data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=self.proc,
        )

        data.output = {"json_field": {"foo": "bar"}}
        data.status = Data.STATUS_DONE
        data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output["json_field"], storage.pk)

    def test_save_storage_file(self):
        """File is loaded and saved to storage"""
        data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=self.proc,
        )
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        data.output = {"json_field": "json.txt"}
        data.status = Data.STATUS_DONE

        json_file = io.StringIO(json.dumps({"foo": "bar"}))

        isfile_mock = MagicMock(return_value=True)
        open_mock = MagicMock(return_value=json_file)
        with patch.object(os.path, "isfile", isfile_mock):
            with patch.object(builtins, "open", open_mock):
                data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output["json_field"], storage.pk)
        self.assertEqual(storage.json, {"foo": "bar"})

    def test_delete_data(self):
        """Orphaned storages are deleted when data object is deleted"""
        storage_referenced_by_both = Storage.objects.create(
            contributor=self.user, json={}
        )

        data_1 = Data.objects.create(contributor=self.contributor, process=self.proc)
        data_1.storages.create(contributor=self.user, json={})
        data_1.storages.add(storage_referenced_by_both)

        data_2 = Data.objects.create(contributor=self.contributor, process=self.proc)
        data_2.storages.add(storage_referenced_by_both)
        data_2.save()

        Storage.objects.create(contributor=self.user, json={})

        self.assertEqual(Data.objects.count(), 2)
        self.assertEqual(Storage.objects.count(), 3)

        # Delete first object.
        data_1.delete()
        self.assertEqual(Data.objects.count(), 1)
        self.assertEqual(Storage.objects.count(), 2)

        # Delete second object.
        data_2.delete()
        self.assertEqual(Data.objects.count(), 0)
        self.assertEqual(Storage.objects.count(), 1)

    def test_storage_manager(self):
        data = Data.objects.create(
            name="Test data", contributor=self.contributor, process=self.proc,
        )

        data.output = {"json_field": {"foo": {"moo": "bar"}}}
        data.status = Data.STATUS_DONE
        data.save()

        # Annotation with specific JSON field.
        storage = Storage.objects.with_json_path(["foo", "moo"])[0]
        self.assertEqual(storage.json_foo_moo, "bar")

        storage = Storage.objects.with_json_path("foo.moo")[0]
        self.assertEqual(storage.json_foo_moo, "bar")

        storage = Storage.objects.with_json_path(["foo", "moo"], field="result")[0]
        self.assertEqual(storage.result, "bar")

        # Extract specific JSON path.
        value = Storage.objects.get_json_path(["foo", "moo"])[0]
        self.assertEqual(value, "bar")

        value = Storage.objects.get_json_path("foo.moo")[0]
        self.assertEqual(value, "bar")


class UtilsTestCase(TestCase):
    def test_render_template(self):
        process_mock = MagicMock(requirements={"expression-engine": "jinja"})
        template = render_template(process_mock, "{{ 1 | increase }}", {})
        self.assertEqual(template, "2")

    def test_render_template_error(self):
        process_mock = MagicMock(requirements={"expression-engine": "jinja"})
        with self.assertRaises(EvaluationError):
            render_template(process_mock, "{{ 1 | missing_increase }}", {})

    def test_hydrate_input_references(self):
        DescriptorSchema.objects.create(
            name="Sample", slug="sample", contributor=self.contributor
        )

        descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor,
            schema=[{"name": "annotation", "type": "basic:string:",},],
        )

        process = Process.objects.create(
            contributor=self.contributor,
            type="data:test:",
            output_schema=[
                {"name": "file", "type": "basic:file:",},
                {"name": "file_list", "type": "list:basic:file:",},
                {"name": "dir", "type": "basic:dir:",},
                {"name": "dir_list", "type": "list:basic:dir:",},
            ],
            entity_type="sample",
            entity_descriptor_schema="sample",
        )

        data = Data.objects.create(
            name="test",
            contributor=self.contributor,
            status=Data.STATUS_ERROR,
            process=process,
            # Workaround for skipping the validation.
            output={
                "file": {"file": "some-file", "refs": ["ref1"]},
                "file_list": [
                    {"file": "some-file", "refs": ["ref2"]},
                    {"file": "another-file"},
                ],
                "dir": {"dir": "some-dir", "refs": ["ref3"]},
                "dir_list": [
                    {"dir": "some-dir", "refs": ["ref4"]},
                    {"dir": "another-dir"},
                ],
            },
            descriptor_schema=descriptor_schema,
            descriptor={"annotation": "my-annotation",},
            size=0,
        )
        data_location = DataLocation.objects.create(subpath="")
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)
        input_schema = [
            {"name": "data", "type": "data:test:",},
        ]
        input_ = {"data": data.pk}
        hydrate_input_references(input_, input_schema)

        path_prefix = data.location.get_path()

        self.assertEqual(
            input_["data"]["__descriptor"], {"annotation": "my-annotation"}
        )
        self.assertEqual(input_["data"]["__type"], "data:test:")
        self.assertEqual(input_["data"]["__id"], data.id)

        self.assertEqual(input_["data"]["file"]["file"].data_id, data.id)
        self.assertEqual(input_["data"]["file"]["file"].file_name, "some-file")
        self.assertEqual(
            str(input_["data"]["file"]["file"]), os.path.join(path_prefix, "some-file")
        )

        self.assertEqual(input_["data"]["file"]["refs"][0].data_id, data.id)
        self.assertEqual(input_["data"]["file"]["refs"][0].file_name, "ref1")
        self.assertEqual(
            str(input_["data"]["file"]["refs"][0]), os.path.join(path_prefix, "ref1")
        )

        self.assertEqual(input_["data"]["file_list"][0]["file"].data_id, data.id)
        self.assertEqual(input_["data"]["file_list"][0]["file"].file_name, "some-file")
        self.assertEqual(
            str(input_["data"]["file_list"][0]["file"]),
            os.path.join(path_prefix, "some-file"),
        )

        self.assertEqual(input_["data"]["file_list"][0]["refs"][0].data_id, data.id)
        self.assertEqual(input_["data"]["file_list"][0]["refs"][0].file_name, "ref2")
        self.assertEqual(
            str(input_["data"]["file_list"][0]["refs"][0]),
            os.path.join(path_prefix, "ref2"),
        )

        self.assertEqual(input_["data"]["file_list"][1]["file"].data_id, data.id)
        self.assertEqual(
            input_["data"]["file_list"][1]["file"].file_name, "another-file"
        )
        self.assertEqual(
            str(input_["data"]["file_list"][1]["file"]),
            os.path.join(path_prefix, "another-file"),
        )

        self.assertEqual(input_["data"]["dir"]["dir"].data_id, data.id)
        self.assertEqual(input_["data"]["dir"]["dir"].file_name, "some-dir")
        self.assertEqual(
            str(input_["data"]["dir"]["dir"]), os.path.join(path_prefix, "some-dir")
        )

        self.assertEqual(input_["data"]["dir"]["refs"][0].data_id, data.id)
        self.assertEqual(input_["data"]["dir"]["refs"][0].file_name, "ref3")
        self.assertEqual(
            str(input_["data"]["dir"]["refs"][0]), os.path.join(path_prefix, "ref3")
        )

        self.assertEqual(input_["data"]["dir_list"][0]["dir"].data_id, data.id)
        self.assertEqual(input_["data"]["dir_list"][0]["dir"].file_name, "some-dir")
        self.assertEqual(
            str(input_["data"]["dir_list"][0]["dir"]),
            os.path.join(path_prefix, "some-dir"),
        )

        self.assertEqual(input_["data"]["dir_list"][0]["refs"][0].data_id, data.id)
        self.assertEqual(input_["data"]["dir_list"][0]["refs"][0].file_name, "ref4")
        self.assertEqual(
            str(input_["data"]["dir_list"][0]["refs"][0]),
            os.path.join(path_prefix, "ref4"),
        )

        self.assertEqual(input_["data"]["dir_list"][1]["dir"].data_id, data.id)
        self.assertEqual(input_["data"]["dir_list"][1]["dir"].file_name, "another-dir")
        self.assertEqual(
            str(input_["data"]["dir_list"][1]["dir"]),
            os.path.join(path_prefix, "another-dir"),
        )

        self.assertEqual(input_["data"]["__entity_name"], "test")



# pylint: disable=missing-docstring,invalid-name
import datetime
from unittest import mock

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.management import call_command
from django.test import override_settings
from django.utils.timezone import get_current_timezone

from guardian.shortcuts import assign_perm
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.elastic.builder import index_builder
from resolwe.test import TestCase

factory = APIRequestFactory()


CUSTOM_SETTINGS = {
    "INSTALLED_APPS": settings.INSTALLED_APPS + ("resolwe.elastic.tests.test_app",),
}


@override_settings(**CUSTOM_SETTINGS)
class IndexViewsetTest(APITestCase, TestCase):
    def setUp(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchIndex
        from .test_app.viewsets import TestViewSet

        super().setUp()

        apps.clear_cache()
        call_command("migrate", verbosity=0, interactive=False)

        index_builder.indexes = [TestSearchIndex()]
        index_builder.register_signals()

        # Prepare users and groups
        user_model = get_user_model()
        self.user_1 = user_model.objects.create(username="user_one")
        self.user_2 = user_model.objects.create(username="user_two")
        group = Group.objects.create(name="group")
        group.user_set.add(self.user_2)

        tzone = get_current_timezone()
        # Prepare test data
        test_obj_1 = TestModel.objects.create(
            name="Object name 1",
            number=43,
            date=datetime.datetime(2018, 1, 1, 0, 0, tzinfo=tzone),
        )
        test_obj_2 = TestModel.objects.create(
            name="Object name 2",
            number=44,
            date=datetime.datetime(2017, 1, 1, 0, 0, tzinfo=tzone),
        )
        test_obj_3 = TestModel.objects.create(
            name="Object name 3",
            number=45,
            date=datetime.datetime(2016, 1, 1, 0, 0, tzinfo=tzone),
        )

        # Assing permissions
        assign_perm("view_testmodel", self.user_1, test_obj_1)
        assign_perm("view_testmodel", group, test_obj_2)
        assign_perm("view_testmodel", AnonymousUser(), test_obj_3)

        # Prepare test viewset
        self.test_viewset = TestViewSet.as_view(actions={"post": "list_with_post",})

    def tearDown(self):
        index_builder.destroy()
        super().tearDown()

    def test_permissions(self):
        # First user
        request = factory.post("", {}, format="json")
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)
        response = sorted(response.data, key=lambda obj: obj["name"])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]["name"], "Object name 1")
        self.assertEqual(response[1]["name"], "Object name 3")

        # Second user
        request = factory.post("", {}, format="json")
        force_authenticate(request, self.user_2)
        response = self.test_viewset(request)
        response = sorted(response.data, key=lambda obj: obj["name"])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]["name"], "Object name 2")
        self.assertEqual(response[1]["name"], "Object name 3")

        # Public user
        request = factory.post("", {}, format="json")
        response = self.test_viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 3")

    def test_without_ordering(self):
        from .test_app.viewsets import TestEmptyOrderingViewSet

        viewset = TestEmptyOrderingViewSet.as_view(actions={"post": "list_with_post",})

        request = factory.post("", {}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)
        response = sorted(response.data, key=lambda obj: obj["name"])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]["name"], "Object name 1")
        self.assertEqual(response[1]["name"], "Object name 3")

    def test_pagination(self):
        request = factory.post("", {"offset": "0", "limit": "1"}, format="json")
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["name"], "Object name 1")

        request = factory.post("", {"offset": "1", "limit": "1"}, format="json")
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["name"], "Object name 3")

    @mock.patch("resolwe.elastic.viewsets.ELASTICSEARCH_SIZE", 1)
    def test_pagination_elasticsearch_size_limit(self):
        request = factory.post("", {"offset": "0", "limit": "1"}, format="json")
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data["results"]), 1)

    def test_custom_filter(self):
        from .test_app.viewsets import TestCustomFieldFilterViewSet

        viewset = TestCustomFieldFilterViewSet.as_view(
            actions={"post": "list_with_post",}
        )

        request = factory.post("", {"name": "1"}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 0)

        request = factory.post("", {"name": "43"}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 1")

    def test_combined_viewset(self):
        from .test_app.viewsets import TestCombinedViewSet

        viewset = TestCombinedViewSet.as_view(actions={"get": "list"})

        # Test database-only access.
        request = factory.get("", {}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]["name"], "Object name 1")
        self.assertEqual(response.data[0]["field_process_type"], "")
        self.assertEqual(response.data[0]["number"], 43)
        self.assertEqual(response.data[1]["name"], "Object name 3")

        # Test combined access.
        request = factory.get("", {"name": "43"}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 0)

        request = factory.get("", {"name": "1"}, format="json")
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 1")
        self.assertEqual(response.data[0]["field_process_type"], "")
        self.assertEqual(response.data[0]["number"], 43)

        # Test combined order.
        request = factory.get("", {"name": "Object", "ordering": "name"})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]["name"], "Object name 1")
        self.assertEqual(response.data[1]["name"], "Object name 3")

        request = factory.get("", {"name": "Object", "ordering": "-name"})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]["name"], "Object name 3")
        self.assertEqual(response.data[1]["name"], "Object name 1")

        # Test order by multiple fields.
        request = factory.get("", {"name": "Object", "ordering": "-name,num"})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]["name"], "Object name 3")
        self.assertEqual(response.data[1]["name"], "Object name 1")

    def _make_request(self, **kwargs):
        request = factory.post("", kwargs, format="json")
        force_authenticate(request, self.admin)
        return self.test_viewset(request)

    def test_filtering_map(self):
        response = self._make_request(name_alias="1")
        self.assertEqual(len(response.data), 1)

    def test_lookup_expressions_number_lt(self):
        response = self._make_request(num__lt="43")
        self.assertEqual(len(response.data), 0)

        response = self._make_request(num__lt="44")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__lt="45")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__lt="100")
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_number_lte(self):
        response = self._make_request(num__lte="43")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__lte="44")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__lte="45")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__lte="100")
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_number_gt(self):
        response = self._make_request(num__gt="0")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gt="43")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__gt="44")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__gt="45")
        self.assertEqual(len(response.data), 0)

        response = self._make_request(num__gt="100")
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_number_gte(self):
        response = self._make_request(num__gte="0")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gte="43")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gte="44")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__gte="45")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__gte="100")
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_date_lt(self):
        response = self._make_request(date__lt="2018-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(date__lt="2017-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__lt="2016-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__lt="2018-01-01T00:00:00.000000")
        self.assertEqual(len(response.data), 2)

    def test_lookup_expressions_date_lte(self):
        response = self._make_request(date__lte="2018-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(date__lte="2017-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__lte="2016-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__lte="2018-01-01T00:00:00.000000")
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_date_gt(self):
        response = self._make_request(date__gt="2018-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 0)

        response = self._make_request(date__gt="2017-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__gt="2016-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__gt="2018-01-01T00:00:00.000000")
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_date_gte(self):
        response = self._make_request(date__gte="2018-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 0)

        response = self._make_request(date__gte="2017-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__gte="2016-03-22T12:23:14.430378")
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__gte="2018-01-01T00:00:00.000000")
        self.assertEqual(len(response.data), 1)

    def test_lookup_expressions_in(self):
        response = self._make_request(num__in="")
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__in="43")
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 1")

        response = self._make_request(num__in="45")
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 3")

        response = self._make_request(num__in="43,45")
        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]["name"], "Object name 1")
        self.assertEqual(response.data[1]["name"], "Object name 3")

    def test_lookup_expressions_exact(self):
        from .test_app.models import TestModel

        tzone = get_current_timezone()
        TestModel.objects.create(
            name="Object name 4",
            number=45,
            date=datetime.datetime(2016, 1, 1, 0, 0, tzinfo=tzone),
            field_process_type="foo:bar:moo:",
        )

        response = self._make_request(field_process_type="foo:bar")
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 4")

        response = self._make_request(field_process_type__exact="foo:bar")
        self.assertEqual(len(response.data), 0)

        response = self._make_request(field_process_type__exact="foo:bar:moo:")
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["name"], "Object name 4")



"""Collection viewset."""
from elasticsearch_dsl.query import Q

from rest_framework import exceptions, mixins, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from resolwe.elastic.composer import composer
from resolwe.elastic.viewsets import ElasticSearchCombinedViewSet
from resolwe.flow.models import Collection
from resolwe.flow.serializers import CollectionSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import update_permission

from ..elastic_indexes import CollectionDocument
from .mixins import (
    ParametersMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)


class CollectionViewSet(
    ElasticSearchCombinedViewSet,
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ParametersMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`Collection` objects."""

    queryset = Collection.objects.all().prefetch_related(
        "descriptor_schema", "contributor"
    )
    serializer_class = CollectionSerializer
    permission_classes = (get_permissions_class(),)
    document_class = CollectionDocument

    filtering_fields = (
        "id",
        "slug",
        "name",
        "name_contains",
        "created",
        "modified",
        "contributor",
        "contributor_name",
        "owners",
        "owners_name",
        "text",
        "tags",
    )
    filtering_map = {
        "name": "name.raw",
        "name_contains": "name.ngrams",
        "contributor": "contributor_id",
        "contributor_name": "contributor_name.ngrams",
        "owners": "owner_ids",
        "owners_name": "owner_names.ngrams",
    }
    ordering_fields = ("id", "created", "modified", "name", "contributor")
    ordering_map = {
        "name": "name.raw",
        "contributor": "contributor_sort",
    }
    ordering = "id"

    def custom_filter_tags(self, value, search):
        """Support tags query."""
        if not isinstance(value, list):
            value = value.split(",")

        filters = [Q("match", **{"tags": item}) for item in value]
        search = search.query("bool", must=filters)

        return search

    def custom_filter_text(self, value, search):
        """Support general query using the 'text' attribute."""
        if isinstance(value, list):
            value = " ".join(value)

        should = [
            Q("match", slug={"query": value, "operator": "and", "boost": 10.0}),
            Q(
                "match",
                **{"slug.ngrams": {"query": value, "operator": "and", "boost": 5.0}},
            ),
            Q("match", name={"query": value, "operator": "and", "boost": 10.0}),
            Q(
                "match",
                **{"name.ngrams": {"query": value, "operator": "and", "boost": 5.0}},
            ),
            Q(
                "match",
                contributor_name={"query": value, "operator": "and", "boost": 5.0},
            ),
            Q(
                "match",
                **{
                    "contributor_name.ngrams": {
                        "query": value,
                        "operator": "and",
                        "boost": 2.0,
                    }
                },
            ),
            Q("match", owner_names={"query": value, "operator": "and", "boost": 5.0}),
            Q(
                "match",
                **{
                    "owner_names.ngrams": {
                        "query": value,
                        "operator": "and",
                        "boost": 2.0,
                    }
                },
            ),
            Q("match", descriptor_data={"query": value, "operator": "and"}),
            Q("match", description={"query": value, "operator": "and"}),
        ]

        # Add registered text extensions.
        for extension in composer.get_extensions(self):
            if hasattr(extension, "text_filter"):
                should += extension.text_filter(value)

        search = search.query("bool", should=should)

        return search

    def set_content_permissions(self, user, obj, payload):
        """Apply permissions to data objects and entities in ``Collection``."""
        for entity in obj.entity_set.all():
            if user.has_perm("share_entity", entity):
                update_permission(entity, payload)

        for data in obj.data.all():
            if user.has_perm("share_data", data):
                update_permission(data, payload)

    def create(self, request, *args, **kwargs):
        """Only authenticated users can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)

    @action(detail=False, methods=["post"])
    def duplicate(self, request, *args, **kwargs):
        """Duplicate (make copy of) ``Collection`` models."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        ids = self.get_ids(request.data)
        queryset = get_objects_for_user(
            request.user, "view_collection", Collection.objects.filter(id__in=ids)
        )
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Collections with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        duplicated = queryset.duplicate(contributor=request.user)

        serializer = self.get_serializer(duplicated, many=True)
        return Response(serializer.data)
